from vowpal_wabbit_next import _core

LabelType = _core.LabelType
SimpleLabel = _core.SimpleLabel
MulticlassLabel = _core.MulticlassLabel
CBLabel = _core.CBLabel
CSLabel = _core.CSLabel
CCBLabel = _core.CCBLabel

CCBExampleType = _core.CCBExampleType
