// Copyright (c) by respective owners including Yahoo!, Microsoft, and
// individual contributors. All rights reserved. Released under a BSD (revised)
// license as described in the file LICENSE.

#pragma once

#define S_EXPLORATION_OK 0
#define E_EXPLORATION_BAD_RANGE 1
#define E_EXPLORATION_PMF_RANKING_SIZE_MISMATCH 2
#define E_EXPLORATION_BAD_PDF 3
#define E_EXPLORATION_BAD_EPSILON 4
